# -*- coding: utf-8 -*-
"""
If flika is loaded using::

    from flika import *

then all of the variables inside global_vars can be accessed as ``g.*``. For instance, to access the 
image inside current window, simply run::

    I = g.currentWindow.image



"""
import os
from multiprocessing import cpu_count
from os.path import expanduser
from qtpy import QtWidgets, QtGui, QtCore
from collections.abc import MutableMapping
import json
from uuid import getnode
from .logger import logger
from .utils.misc import get_location

__all__ = ['m', 'Settings', 'menus', 'alert', 'windows', 'traceWindows', 'currentWindow', 'currentTrace', 'clipboard']

class Settings(MutableMapping): #http://stackoverflow.com/questions/3387691/python-how-to-perfectly-override-a-dict
    """
    All of flika's settings are stored in this object, which is designed to act like a dictionary. When any value in
    this object is changed, this object is converted to json and saved in ``~/.FLIKA/settings.json``. When flika is
    restarted, this settings object is populated with values from this file. Settings can be accessed using 
    `g.settings`. 


    """
    initial_settings = {'filename': None, 
                        'internal_data_type': 'float64',
                        'multiprocessing': True, 
                        'multipleTraceWindows': False, 
                        'mousemode': 'rectangle', 
                        'show_windows': True, 
                        'recent_scripts': [],
                        'recent_files': [],
                        'nCores':cpu_count(),
                        'debug_mode': False,
                        'point_color': '#ff0000',
                        'point_size': 5,
                        'roi_color': '#ffff00',
                        'rect_width': 5,
                        'rect_height': 5,
                        'show_all_points': False,
                        'default_roi_on_click': False}

    def __init__(self):
        self.settings_file = os.path.join(expanduser("~"), '.FLIKA', 'settings.json' )
        self.d = Settings.initial_settings
        self.load()

    def __getitem__(self, item):
        try:
            self.d[item]
        except KeyError:
            self.d[item] = Settings.initial_settings[item] if item in Settings.initial_settings else None
        return self.d[item]

    def __setitem__(self, key, item):
        self.d[key] = item
        self.save()

    def __delitem__(self, key):
        del self.d[key]

    def __iter__(self):
        return iter(self.d)

    def __len__(self):
        return len(self.d)

    def __contains__(self, item):
        return item in self.d

    def save(self):
        """ Save settings file. The file is stored in ``~/.FLIKA/settings.json`` """
        if not os.path.exists(os.path.dirname(self.settings_file)):
            os.makedirs(os.path.dirname(self.settings_file))
        with open(self.settings_file, 'w') as fp:
            json.dump(self.d, fp, indent=4)

    def load(self):
        """ Load settings file. The file is stored in ``~/.FLIKA/settings.json`` """
        if not os.path.exists(self.settings_file):
            print('No settings file found. Creating settings file.')
            self.save()
        try:
            with open(self.settings_file, 'r') as fp:
                d = json.load(fp)
            d = {k: d[k] for k in d if d[k] is not None}
            self.d.update(d)
        except Exception as e:
            msg = "Failed to load settings file. {}\nDefault settings restored.".format(e)
            logger.info(msg)
            print(msg)
            self.save()
        self.d['mousemode'] = 'rectangle'  # don't change initial mousemode
        self._load_user_information()

    def _load_user_information(self):
        if 'user_information' not in self.d:
            self.d['user_information'] = {}
        if 'UUID' not in self.d['user_information'] or self.d['user_information']['UUID'] is None:
            self.d['user_information']['UUID'] = getnode()
        if 'location' not in self.d['user_information'] or self.d['user_information']['location'] is None:
            self.d['user_information']['location'] = get_location()


    def setmousemode(self, mode):
        self['mousemode']=mode

    def setMultipleTraceWindows(self, f):
        self['multipleTraceWindows'] = f

    def setInternalDataType(self, dtype):
        self['internal_data_type'] = dtype
        print('Changed data_type to {}'.format(dtype))


def messageBox(title, text, buttons=QtWidgets.QMessageBox.Ok, icon=QtWidgets.QMessageBox.Information):
    m.messagebox = QtWidgets.QMessageBox(icon, title, text, buttons)
    m.messagebox.setWindowIcon(m.windowIcon())
    m.messagebox.show()
    while m.messagebox.isVisible(): QtWidgets.QApplication.instance().processEvents()
    return m.messagebox.result()
    

def setConsoleVisible(v):
    from ctypes import windll
    GetConsoleWindow = windll.kernel32.GetConsoleWindow
    console_window_handle = GetConsoleWindow()
    ShowWindow = windll.user32.ShowWindow
    ShowWindow(console_window_handle, v)


class SetCurrentWindowSignal(QtWidgets.QWidget):
    sig = QtCore.Signal()

    def __init__(self,parent):
        QtWidgets.QWidget.__init__(self,parent)
        self.hide()


def alert(msg, title="flika - Alert"):
    """Creates a popup that alerts the user.
    """

    print('\nAlert: ' + msg)
    msgbx = QtWidgets.QMessageBox(m)
    msgbx.setIcon(QtWidgets.QMessageBox.Information)
    msgbx.setText(msg)
    msgbx.setWindowTitle(title)
    msgbx.show()
    m.statusBar().showMessage(msg)
    desktopSize = QtWidgets.QDesktopWidget().screenGeometry()
    top = (desktopSize.height() / 2) - (msgbx.size().height() / 2)
    left = (desktopSize.width() / 2) - (msgbx.size().width() / 2)
    msgbx.move(left, top)
    dialogs.append(msgbx)


settings = Settings()
m = None  #: The main window.
menus = []
windows = [] #: list of :class:`windows<flika.window.Window>`: All of the windows that have been created and have not yet been closed.
traceWindows = [] #: list of :class:`TraceFigs<flika.tracefig.TraceFig>`: All of the TraceFigs that are open.
dialogs = []
currentWindow = None #: :class:`window <flika.window.Window>`: The window that is currently selected
currentTrace = None #: :class:`tracefigs<flika.tracefig.TraceFig>`: The tracefig that is currently selected
clipboard = None









# -*- coding: utf-8 -*-
import numpy as np
from qtpy import QtWidgets, QtCore, QtGui
from .. import global_vars as g
from ..window import Window
from .BaseProcess import BaseProcess, WindowSelector, MissingWindowError, CheckBox

__all__ = ['split_channels']


class Split_channels(BaseProcess):
    """ split_channels(keepSourceWindow=False)
    This splits the color channels in a Window

    Returns:
        list of new Windows
    """

    def __init__(self):
        super().__init__()

    def gui(self):
        self.gui_reset()
        super().gui()

    def __call__(self, keepSourceWindow=False):
        self.start(keepSourceWindow)
        newWindows = []
        if not self.oldwindow.metadata['is_rgb']:
            g.alert('Cannot split channels, no colors detected.')
            return None
        nChannels = self.tif.shape[-1]
        for i in range(nChannels):
            newtif = self.tif[..., i]
            name = self.oldname + ' - Channel ' + str(i)
            newWindow = Window(newtif, name, self.oldwindow.filename)
            newWindows.append(newWindow)
        if keepSourceWindow is False:
            self.oldwindow.close()
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        return newWindows


split_channels = Split_channels()


import logging, sys
from logging import NullHandler
from .flika import start_flika
from . import global_vars as g
from .window import Window
from .version import __version__
from .process import *
from .roi import open_rois

# import modules that are commonly used inside flika
import numpy as np
from pyqtgraph import plot, show
import pyqtgraph as pg
import pyqtgraph.opengl as gl


logging.getLogger('flika').addHandler(NullHandler())
def handle_exception(exc_type, exc_value, exc_traceback):
    try:
        logging.getLogger('flika').warn(str(exc_type) + "\n" + str(exc_value) + " " + str(exc_traceback.tb_frame.f_code))
    except:
        logging.getLogger('flika').warn(str(exc_type) + "\n" + str(exc_value))

    sys.__excepthook__(exc_type, exc_value, exc_traceback)
sys.excepthook = handle_exception


from logging import getLogger, basicConfig
import os
LOG_DIR = os.path.join(os.path.expanduser("~"), '.FLIKA', 'log')
if not os.path.exists(LOG_DIR):
    os.makedirs(LOG_DIR)
LOG_PATH = os.path.join(LOG_DIR, 'FLIKALOG.log')
basicConfig(filename=LOG_PATH)
logger = getLogger("flika")


import sys, os
import optparse

from .. import global_vars as g
from ..window import Window
import numpy as np
import time
import pytest
from ..roi import makeROI
import pyqtgraph as pg
from qtpy import QtGui
from qtpy.QtWidgets import qApp
im = np.random.random([120, 90, 90])
	
class TestWindow():
	def setup_method(self, obj):
		self.win1 = Window(im)

	def teardown_method(self):
		self.win1.close()

	def test_link(self):
		win2 = Window(im)
		self.win1.link(win2)
		self.win1.setIndex(100)
		assert win2.currentIndex == self.win1.currentIndex, "Linked windows failure"
		win2.setIndex(50)
		assert self.win1.currentIndex == win2.currentIndex, "Linked windows failure"
		win2.unlink(self.win1)
		self.win1.setIndex(100)
		assert win2.currentIndex != self.win1.currentIndex, "Unlinked windows failure"
		win2.setIndex(50)
		assert self.win1.currentIndex != win2.currentIndex, "Unlinked windows failure"
		win2.close()
		assert len(self.win1.linkedWindows) == 0, "Closed window is not unlinked"

	def test_timeline(self):
		self.win1.imageview.setImage(im[0])
		assert self.win1.imageview.ui.roiPlot.isVisible() == False
		self.win1.imageview.setImage(im)
		assert self.win1.imageview.ui.roiPlot.isVisible() == True

class ROITest():
	TYPE=None
	POINTS=[]
	MASK=[]

	def setup_method(self):
		for i in range(3):
			try:
				self.win1 = Window(self.img)
				break
			except RuntimeError:
				pass
		if not hasattr(self, 'win1'):
			raise Exception("Unable to create Window due to RuntimeError")

		self.changed = False
		self.changeFinished = False
		self.roi = makeROI(self.TYPE, self.POINTS, window=self.win1)
		self.roi.sigRegionChanged.connect(self.preChange)
		self.roi.sigRegionChangeFinished.connect(self.preChangeFinished)
		self.check_placement()

	def preChange(self):
		self.changed = True
	
	def preChangeFinished(self):
		self.changeFinished = True

	def checkChanged(self):
		assert self.changed, "Change signal was not sent"
		self.changed = False
	
	def checkChangeFinished(self):
		assert self.changeFinished, "ChangeFinished signal was not sent"
		self.changeFinished = False
	
	def teardown_method(self, func):
		#if self.roi is not None:
		#	self.roi.delete()
		#assert self.roi not in self.win1.rois, "ROI deleted but still in window.rois"
		for roi in self.win1.rois:
			roi.delete()
		self.win1.close()
		from .conftest import fa
		fa().clear()
		pg.ViewBox.AllViews.clear()
		pg.ViewBox.NamedViews.clear()



	def check_placement(self, mask=None, points=None):
		if mask is None:
			mask = self.MASK
		if points is None:
			points = self.POINTS
		assert np.array_equal(self.roi.getMask(), mask), "Mask differs on creation. %s != %s" % (self.roi.getMask(), self.MASK)
		assert np.array_equal(self.roi.pts, points), "pts differs on creation. %s != %s" % (self.roi.pts, self.POINTS)
		assert np.array_equal(self.roi.getPoints(), points), "getPoints differs on creation. %s != %s" % (self.roi.getPoints(), self.POINTS)

	def check_similar(self, other):
		mask1 = self.roi.getMask()
		mask2 = other.getMask()
		assert np.array_equal(mask1, mask2), "Mask differs on creation. %s != %s" % (mask1, mask2)
		assert np.array_equal(self.roi.pts, other.pts), "pts differs on creation. %s != %s" % (self.roi.pts, self.POINTS)
		assert np.array_equal(self.roi.getPoints(), other.getPoints()), "getPoints differs on creation. %s != %s" % (self.roi.getPoints(), other.getPoints())

	def test_copy(self):
		self.roi.copy()
		roi1 = self.win1.paste()
		assert roi1 == None and len(self.win1.rois) == 1, "Copying ROI to same window"
		
		w2 = Window(self.img)
		roi2 = w2.paste()
		assert self.roi in roi2.linkedROIs and roi2 in self.roi.linkedROIs, "Linked ROI on paste"
		self.check_similar(roi2)

		self.roi.translate([1, 2])

		self.checkChanged()
		self.checkChangeFinished()

		self.check_similar(roi2)

		w2.close()
		self.win1.setAsCurrentWindow()

	def test_export_import(self):
		s = str(self.roi)
		self.roi.window.save_rois("test.txt")
		from ..roi import open_rois
		rois = open_rois('test.txt')
		assert len(rois) == 1, "Import ROI failure"
		roi = rois[0]
		os.remove('test.txt')
		self.check_similar(roi)
		roi.delete()
	
	def test_plot(self):
		self.roi.unplot()
		trace = self.roi.plot()
		if trace is None:
			assert self.win1.image.ndim == 4, "Trace failed on non-4D image"
			return
		assert self.roi.traceWindow != None, "ROI plotted, roi traceWindow set"
		assert g.currentTrace == trace, "ROI plotted, but currentTrace is still None"
		ind = trace.get_roi_index(self.roi)
		if trace.rois[ind]['p1trace'].opts['pen'] != None:
			assert trace.rois[ind]['p1trace'].opts['pen'].color().name() == self.roi.pen.color().name(), "Color not changed. %s != %s" % (trace.rois[ind]['p1trace'].opts['pen'].color().name(), self.roi.pen.color().name())
		self.roi.unplot()
		assert self.roi.traceWindow == None, "ROI unplotted, roi traceWindow cleared"
		assert g.currentTrace == None, "ROI unplotted, currentTrace cleared"
		g.settings['multipleTraceWindows'] = False
	
	def test_translate(self):
		path = self.roi.pts.copy()
		self.roi.translate([2, 1])

		self.checkChanged()
		self.checkChangeFinished()

		assert len(self.roi.pts) == len(path), "roi size change on translate"
		assert [i + [2, 1] in path for i in self.roi.pts], "translate applied to mask"
		self.roi.translate([-2, -1])

		self.checkChanged()
		self.checkChangeFinished()

	def test_plot_translate(self):
		trace = self.roi.plot()
		if trace is None:
			assert self.win1.image.ndim == 4, "Trace failed on non-4D image"
			return
		assert self.roi.traceWindow != None, "ROI plotted, roi traceWindow set. %s should not be None" % (self.roi.traceWindow)
		assert g.currentTrace == self.roi.traceWindow, "ROI plotted, currentTrace not set. %s != %s" % (g.currentTrace, self.roi.traceWindow) 
		ind = trace.get_roi_index(self.roi)

		traceItem = trace.rois[ind]['p1trace']
		yData = traceItem.yData.copy()
		self.roi.translate(2, 1)
		assert not np.array_equal(yData, traceItem.yData), "Translated ROI yData compare %s != %s" % (yData, traceItem.yData)
		
		self.checkChanged()
		self.checkChangeFinished()

		self.roi.translate(-2, -1)
		assert np.array_equal(yData, traceItem.yData), "Translated back ROI yData compare %s != %s" % (yData, traceItem.yData)

		self.checkChanged()
		self.checkChangeFinished()

		self.roi.unplot()

	def test_change_color(self):
		self.roi.unplot()
		color = QtGui.QColor('#ff00ff')
		self.roi.plot()
		self.roi.colorSelected(color)
		
		self.checkChangeFinished()

		assert self.roi.pen.color().name() == color.name(), "Color not changed. %s != %s" % (self.roi.pen.color().name(), color.name())
		self.roi.unplot()
	
	
	def test_translate_multiple(self):
		translates = [[5, 0], [0, 5], [-5, 0], [0, -5]]
		self.roi.copy()
		w2 = Window(self.img)
		roi2 = w2.paste()
		roi2.colorSelected(QtGui.QColor(0, 255, 130))
		self.roi.plot()
		roi2.plot()
		for i in range(20 * len(translates)):
			tr = translates[i % len(translates)]
			self.roi.translate(*tr)
			self.checkChanged()
			self.checkChangeFinished()
			#time.sleep(.02)
			#qApp.processEvents()

		w2.close()
		self.roi.draw_from_points(self.POINTS)
	
	def test_resize_multiple(self):
		if len(self.roi.getHandles()) == 0:
			return
		translates = [[1, 0], [0, 1], [-1, 0], [0, -1]]
		self.roi.copy()
		w2 = Window(self.img)
		roi2 = w2.paste()
		roi2.colorSelected(QtGui.QColor(0, 255, 130))
		self.roi.plot()
		roi2.plot()

		for h in self.roi.getHandles():
			h._updateView()
			pos = h.viewPos()
			for i in range(4 * len(translates)):
				tr = translates[i % len(translates)]
				self.roi.movePoint(h, [pos.x() + tr[0], pos.y() + tr[1]])
				self.checkChanged()
				self.checkChangeFinished()
				self.check_similar(roi2)
				#time.sleep(.02)
				#qApp.processEvents()

		w2.close()
		self.roi.draw_from_points(self.POINTS)

class ROI_Rectangle(ROITest):
	TYPE = "rectangle"
	POINTS = [[3, 2], [2, 5]]
	MASK = [[3, 4, 3, 4, 3, 4, 3, 4, 3, 4], [2, 2, 3, 3, 4, 4, 5, 5, 6, 6]]

	def test_crop(self):
		w2 = self.roi.crop()
		bound = self.roi.boundingRect()
		mask = self.roi.getMask()
		w, h = np.ptp(mask, 1) + [1, 1]

		assert w == bound.width() and h == bound.height(), "Croppped image different size (%s, %s) != (%s, %s)" % (bound.width(), bound.height(), w, h)
		w2.close()

	def test_resize(self):
		self.roi.scale([1, 1.2])

		self.checkChanged()
		self.checkChangeFinished()
		
		points = [self.POINTS[0], [2, 6]]
		mask = [[3, 4, 3, 4, 3, 4, 3, 4, 3, 4, 3, 4], [2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7]]
		self.check_placement(points=points, mask=mask)

		self.roi.draw_from_points(self.POINTS)

class ROI_Line(ROITest):
	TYPE="line"
	POINTS = [[3, 2], [8, 4]]
	MASK = [[3, 4, 5, 6, 7, 8], [2, 2, 3, 3, 4, 4]]

	def test_kymograph(self):
		self.roi.update_kymograph()
		if self.roi.kymograph is None and self.win1.image.ndim != 3:
			return
		kymo = self.roi.kymograph
		assert kymo.image.shape[1] == self.win1.image.shape[0]
		kymo.close()
		self.win1.setAsCurrentWindow()

	def test_move_handle(self):
		self.roi.movePoint(0, [3, 4])

		self.checkChanged()
		self.checkChangeFinished()
		
		newMask = [[3, 4, 5, 6, 7, 8], [4, 4, 4, 4, 4, 4]]
		assert np.array_equal(self.roi.getMask(), newMask)
		assert np.array_equal(self.roi.getPoints(), [[3, 4], [8, 4]])
		self.roi.movePoint(0, [3, 2])

		self.checkChanged()
		self.checkChangeFinished()


class ROI_Freehand(ROITest):
	TYPE="freehand"
	POINTS = [3, 2], [5, 6], [2, 4]
	MASK = [[3, 3, 3, 4, 4], [2, 3, 4, 4, 5]]

	def test_translate_multiple(self):
		pass


class ROI_Rect_Line(ROITest):
	TYPE="rect_line"
	POINTS = [[3, 2], [5, 4], [4, 8]]
	MASK = [[3, 4, 5, 5, 5, 4, 4, 4], [2, 3, 4, 4, 5, 6, 7, 8]]

	
	def test_extend(self):
		self.roi.extend(9, 2)

		self.checkChanged()
		self.checkChangeFinished()

		self.roi.removeSegment(len(self.roi.lines)-1)
	
	def test_plot_translate(self):
		pass

	def test_translate(self):
		pass

	def test_translate_multiple(self):
		pass

	def test_kymograph(self):
		self.roi.update_kymograph()
		if self.roi.kymograph is None and self.win1.image.ndim != 3:
			return
		kymo = self.roi.kymograph
		assert kymo.image.shape[1] == self.win1.image.shape[0]
		
		self.roi.setWidth(3)
		assert kymo.image.shape[1] == self.win1.image.shape[0]

		kymo.close()
		self.win1.setAsCurrentWindow()

	def test_copy(self):
		self.roi.copy()
		roi1 = self.win1.paste()
		assert roi1 == None and len(self.win1.rois) == 1, "Copying ROI to same window"
		
		w2 = Window(self.img)
		roi2 = w2.paste()
		assert self.roi in roi2.linkedROIs and roi2 in self.roi.linkedROIs, "Linked ROI on paste"
		self.check_similar(roi2)

		self.roi.lines[0].movePoint(0, [1, 2])

		self.checkChanged()
		self.checkChangeFinished()

		self.check_similar(roi2)

		w2.close()
		self.win1.setAsCurrentWindow()


class Test_Rectangle_2D(ROI_Rectangle):
	img = np.random.random([20, 20])
class Test_Rectangle_3D(ROI_Rectangle):
	img = np.random.random([10, 20, 20])
class Test_Rectangle_4D(ROI_Rectangle):
	img = np.random.random([10, 20, 20, 3])

class Test_Line_2D(ROI_Line):
	img = np.random.random([20, 20])
class Test_Line_3D(ROI_Line):
	img = np.random.random([10, 20, 20])
class Test_Line_4D(ROI_Line):
	img = np.random.random([10, 20, 20, 3])

class Test_Freehand_2D(ROI_Freehand):
	img = np.random.random([20, 20])
class Test_Freehand_3D(ROI_Freehand):
	img = np.random.random([10, 20, 20])
class Test_Freehand_4D(ROI_Freehand):
	img = np.random.random([10, 20, 20, 3])

class Test_Rect_Line_2D(ROI_Rect_Line):
	img = np.random.random([20, 20])
class Test_Rect_Line_3D(ROI_Rect_Line):
	img = np.random.random([10, 20, 20])
class Test_Rect_Line_4D(ROI_Rect_Line):
	img = np.random.random([10, 20, 20, 3])




class TestTracefig():
	def setup_method(self):
		self.w1 = Window(im)
		self.rect = makeROI('rectangle', [[3, 2], [4, 5]], window=self.w1)
		self.trace = self.rect.plot()

	def teardown_method(self):
		self.w1.close()

	def test_plotting(self):
		self.trace.indexChanged.emit(20)
		assert self.w1.currentIndex == 20, "trace indexChanged"

	def test_export(self):
		self.rect.window.save_rois('tempROI.txt')
		t = open('tempROI.txt').read()
		assert t == 'rectangle\n3 2\n4 5\n'
		os.remove('tempROI.txt')



#!/usr/bin/env python
"""
Commands to upload to pypi:

python setup.py sdist bdist_wheel
twine upload dist/*
"""
from setuptools import setup, find_packages
from distutils.core import Command
from setuptools.command.install import install
import platform
import os
import sys

with open('flika/version.py') as infile:
    __version__ = '0.0.0'
    exec(infile.read())  # This sets the __version__ variable.

with open('README.rst') as readme:
    LONG_DESCRIPTION = readme.read()
    LONG_DESCRIPTION = LONG_DESCRIPTION.replace('.. image:: flika/docs/_static/img/flika_screencapture.gif', '')

cmdclass = {}


class PyTest(Command):

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        self.pytest_args = ""

    def finalize_options(self):
        pass

    def run(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args + ' flika')
        sys.exit(errno)


cmdclass['test'] = PyTest

entry_points = """
[console_scripts]
flika = flika.flika:exec_
flika_post_install = flika.flika:post_install
"""

setup_requires = ['numpy', 'scipy']

# must have numpy and scipy installed already
install_requires = [
      'pandas>=0.14',
      'matplotlib>=1.4',
      'pyqtgraph>=0.9',
      'PyQt5>=5.8.0',
      'qtpy>=1.1',
      'setuptools>=1.0',
      'scikit-image',
      'scikit-learn',
      'ipython>=1.0',
      'ipykernel',
      'qtconsole',
      'pyopengl',
      'requests',
      'nd2reader',
      'markdown']

extras_require = {
    ':sys_platform == "win32"': ['winshell', 'pypiwin32']
}

if sys.platform == 'win32':
    install_requires += ['winshell', 'pypiwin32']



setup(name='flika',
      version=__version__,
      cmdclass=cmdclass,
      description='An interactive image processing program for biologists written in Python.',
      long_description=LONG_DESCRIPTION,
      author='Kyle Ellefsen, Brett Settle',
      author_email='kyleellefsen@gmail.com',
      url='http://flika-org.github.io',
      setup_requires=setup_requires,
      install_requires=install_requires,
      extras_require=extras_require,
      license='MIT',
      classifiers=[
          'Intended Audience :: Science/Research',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering :: Visualization',
          ],
      packages=find_packages(),
      entry_points=entry_points,
      include_package_data=True,
      package_data={'gui': ['*.ui'],
                    'images': ['*.ico', '*.png', '*.txt', '*.tif']})












import os
import random
import numpy as np
import platform
import json
import requests
import time
from qtpy import uic, QtGui, QtWidgets
__all__ = ['nonpartial', 'setConsoleVisible', 'load_ui', 'random_color', 'save_file_gui', 'open_file_gui', 'get_location']

def nonpartial(func, *args, **kwargs):
    """Like functools.partial, this returns a function which, when called, calls
    ``func(*args, **kwargs)``.

    Unlike functools.partial, extra arguments passed to the returned function
    are *not* passed to the input function.
    """
    def result(*a, **k):
        return func(*args, **kwargs)

    return result

def setConsoleVisible(v):
    """ Set visibility of the console when running flika. This only works on windows systems
    when running flika as a standalone process and may be removed soon.

    Args:
        v (bool): True to show console, False to hide

    """
    if platform.system() == 'Windows':
        from ctypes import windll
        GetConsoleWindow = windll.kernel32.GetConsoleWindow
        console_window_handle = GetConsoleWindow()
        ShowWindow = windll.user32.ShowWindow
        ShowWindow(console_window_handle, v)
    else:
        print('Displaying the console on non-windows systems not yet supported')

def random_color():
    """ Generate an QColor that is bright enough to see on a black and white background
    
    Returns:
        QtGui.QColor: randomly generated color object
    """ 
    colors = [(165,42,42), (178,34,34), (220,20,60), (255,0,0), (255,99,71), (255,127,80), (205,92,92), (240,128,128), (233,150,122), 
    (250,128,114), (255,160,122), (255,69,0), (255,140,0), (255,165,0), (255,215,0), (184,134,11), (218,165,32), (238,232,170), (240,230,140),
    (128,128,0), (255,255,0), (154,205,50), (85,107,47), (107,142,35), (124,252,0), (127,255,0), (173,255,47), (34,139,34), (0,255,0),
    (50,205,50), (144,238,144), (152,251,152), (0,250,154), (0,255,127), (46,139,87), (102,205,170), (60,179,113), (32,178,170),
    (0,128,128), (0,139,139), (0,255,255), (0,255,255), (0,206,209), (64,224,208), (72,209,204), (175,238,238), (127,255,212), (176,224,230),
    (95,158,160), (70,130,180), (100,149,237), (0,191,255), (30,144,255), (173,216,230), (135,206,235), (135,206,250), 
    (138,43,226), (75,0,130), (106,90,205), (123,104,238), (147,112,219), (139,0,139), (148,0,211),
    (153,50,204), (186,85,211), (128,0,128), (216,191,216), (221,160,221), (238,130,238), (255,0,255), (218,112,214), (199,21,133), (219,112,147),
    (255,20,147), (255,105,180), (255,182,193), (255,192,203), (250,235,215), (245,245,220), (255,228,196), (255,235,205), (245,222,179), (255,248,220),
    (255,250,205), (250,250,210), (255,255,224), (139,69,19), (160,82,45), (210,105,30), (205,133,63), (244,164,96), (222,184,135), (210,180,140),
    (188,143,143), (255,228,181), (255,222,173), (255,218,185), (255,228,225), (255,240,245), (250,240,230), (253,245,230), (255,239,213), (255,245,238),
    (245,255,250), (176,196,222), (230,230,250), (240,248,255), (248,248,255), (240,255,240)]
    ind = np.random.choice(range(len(colors)))
    return QtGui.QColor(*colors[ind])

def load_ui(path, parent=None, directory=None):
    """Load a .ui file

    Args:
        path (str): Name of ui file to load
        parent (QtCore.QObject): Object to use as the parent of this widget

    Returns:
        QtWidgets.QWidget: The widget created from the specified ui file
    """

    if directory is not None:
        full_path = os.path.join(directory, path)
    else:
        full_path = os.path.abspath(path)

    if not os.path.exists(full_path) and 'site-packages.zip' in full_path:
        # Workaround for Mac app
        full_path = os.path.join(full_path.replace('site-packages.zip', 'flika'))

    return uic.loadUi(full_path, parent)

def save_file_gui(prompt="Save file", directory=None, filetypes=''):
    """ File dialog for saving a new file, isolated to handle tuple/string return value
    
    Args:
        prompt (str): string to display at the top of the window
        directory (str): initial directory to save the file to
        filetypes (str): argument for filtering file types separated by ;; (*.png) or (Images *.png);;(Other *.*)
    
    Returns:
        str: the file path selected, or empty string if canceled
    """
    from .. import global_vars as g
    if directory is None or directory == '':
        filename = g.settings['filename']
        try:
            directory = os.path.dirname(filename)
        except:
            directory = None
    if directory is None:
        filename = QtWidgets.QFileDialog.getSaveFileName(g.m, prompt, filter=filetypes)
    else:
        filename = QtWidgets.QFileDialog.getSaveFileName(g.m, prompt, directory, filter=filetypes)
    if isinstance(filename, tuple):
        filename, ext = filename
        if ext and '.' not in filename:
            filename += '.' + ext.rsplit('.')[-1]
    if filename is None or str(filename) == '':
        g.m.statusBar().showMessage('Save Cancelled')
        return None
    else:
        return str(filename)


def open_file_gui(prompt="Open File", directory=None, filetypes=''):
    """ File dialog for opening an existing file, isolated to handle tuple/string return value
    
    Args:
        prompt (str): string to display at the top of the window
        directory (str): initial directory to open
        filetypes (str): argument for filtering file types separated by ;; (*.png) or (Images *.png);;(Other *.*)
    
    Returns:
        str: the file (path+file+extension) selected, or None
    """
    from .. import global_vars as g
    filename = None
    if directory is None:
        filename = g.settings['filename']
        try:
            directory = os.path.dirname(filename)
        except:
            directory = None
    if directory is None or filename is None:
        filename = QtWidgets.QFileDialog.getOpenFileName(g.m, prompt, '', filetypes)
    else:
        filename = QtWidgets.QFileDialog.getOpenFileName(g.m, prompt, filename, filetypes)
    if isinstance(filename, tuple):
        filename, ext = filename
        if ext and '.' not in filename:
            filename += '.' + ext.rsplit('.')[-1]
    if filename is None or str(filename) == '':
        g.m.statusBar().showMessage('No File Selected')
        return None
    else:
        return str(filename)

def send_error_report(email, report):
    """Log error reports to the flikarest.pythonanywhere REST API to be stored in a SQL database

    Parameters:
        email (str): Email of the user, optionally entered for response
        report (str): The complete error that occurred

    Returns:
        Request Response if the API was reached, or None if connection failed. Use response.status==200 to check success

    """
    from .. import global_vars as g
    address = g.settings['user_information']['UUID']
    location = g.settings['user_information']['location']

    kargs = {'address': address, 'email': email, 'report': report, 'location': location}
    try:
        r = requests.post("http://flikarest.pythonanywhere.com/reports/submit", data=kargs)
    except requests.exceptions.ConnectionError:
        return None
    return r

def send_user_stats():
    """Log user information to the flikarest.pythonanywhere REST API to be stored in a SQL database

    Currently uses the get_node to get a UUID for the machine, and the IP address location API to get a rough 
    'city, region, country' location, which are only retrieved once and stored in global settings.

    Returns:
        Request Response if the API was reached, or None if connection failed. Use response.status==200 to check success

    """
    from .. import global_vars as g
    address = g.settings['user_information']['UUID']
    location = g.settings['user_information']['location']

    kargs = {'address': address, 'location': location}
    try:
        r = requests.post("http://flikarest.pythonanywhere.com/user_stats/log_user", data=kargs)
    except requests.exceptions.ConnectionError:
        return None
    return r

def get_location():
    """Call to a location REST API that uses the current IP address to get a string representation of the geolocation

    Returns:
        str: the 'city, region, country' location of the current IP address. Or None if the connection fails
    """
    send_url = 'http://freegeoip.net/json'
    try:
        r = requests.get(send_url)
    except (requests.exceptions.ConnectionError, Exception):
        return None
    j = json.loads(r.text)
    location = "{}, {}, {}".format(j['city'], j['region_name'], j['country_name'])
    return location


# -*- coding: utf-8 -*-
from .stacks import *
from .math_ import *
from .filters import *
from .binary import *
from .roi import *
from .measure import *
from .color import *
from .overlay import *
from .file_ import *

def setup_menus():
    from .. import global_vars as g
    if len(g.menus) > 0:
        print("flika menubar already initialized.")
        return
    from qtpy import QtGui, QtWidgets
    imageMenu = QtWidgets.QMenu("Image")
    processMenu = QtWidgets.QMenu("Process")

    def addAction(menu, name, trigger):
        menu.addAction(QtWidgets.QAction(name, menu, triggered=trigger))

    stacksMenu = imageMenu.addMenu("Stacks")

    addAction(stacksMenu, "Duplicate", duplicate)
    addAction(stacksMenu, "Generate Random Image", generate_random_image.gui)
    addAction(stacksMenu, "Trim Frames", trim.gui)
    addAction(stacksMenu, "Deinterlace", deinterleave.gui)
    addAction(stacksMenu, "Z Project", zproject.gui)
    addAction(stacksMenu, "Pixel Binning", pixel_binning.gui)
    addAction(stacksMenu, "Frame Binning", frame_binning.gui)
    addAction(stacksMenu, "Resize", resize.gui)
    addAction(stacksMenu, "Concatenate Stacks", concatenate_stacks.gui)
    addAction(stacksMenu, "Change Data Type", change_datatype.gui)

    colorMenu = imageMenu.addMenu("Color")
    addAction(colorMenu, "Split Channels", split_channels.gui)

    addAction(imageMenu, "Measure", measure.gui)
    addAction(imageMenu, "Set Value", set_value.gui)
    overlayMenu = imageMenu.addMenu("Overlay")
    addAction(overlayMenu, "Background", background.gui)
    addAction(overlayMenu, "Timestamp", time_stamp.gui)
    addAction(overlayMenu, "Scale Bar", scale_bar.gui)

    binaryMenu = processMenu.addMenu("Binary")
    mathMenu = processMenu.addMenu("Math")
    filtersMenu = processMenu.addMenu("Filters")
    processMenu.addAction(QtWidgets.QAction("Image Calculator", processMenu, triggered=image_calculator.gui))

    addAction(binaryMenu, "Threshold", threshold.gui)
    addAction(binaryMenu, "Adaptive Threshold", adaptive_threshold.gui)
    addAction(binaryMenu, "Canny Edge Detector", canny_edge_detector.gui)
    binaryMenu.addSeparator()
    addAction(binaryMenu, "Logically Combine", logically_combine.gui)
    addAction(binaryMenu, "Remove Small Blobs", remove_small_blobs.gui)
    addAction(binaryMenu, "Binary Erosion", binary_erosion.gui)
    addAction(binaryMenu, "Binary Dilation", binary_dilation.gui)
    addAction(binaryMenu, "Generate ROIs", generate_rois.gui)

    addAction(mathMenu, "Multiply", multiply.gui)
    addAction(mathMenu, "Divide", divide.gui)
    addAction(mathMenu, "Subtract", subtract.gui)
    addAction(mathMenu, "Power", power.gui)
    addAction(mathMenu, "Ratio By Baseline", ratio.gui)
    addAction(mathMenu, "Absolute Value", absolute_value.gui)
    addAction(mathMenu, "Subtract Trace", subtract_trace.gui)
    addAction(mathMenu, "Divide Trace", divide_trace.gui)

    addAction(filtersMenu, "Gaussian Blur", gaussian_blur.gui)
    addAction(filtersMenu, "Difference of Gaussians", difference_of_gaussians.gui)
    filtersMenu.addSeparator()
    addAction(filtersMenu, "Butterworth Filter", butterworth_filter.gui)
    addAction(filtersMenu, "Mean Filter", mean_filter.gui)
    addAction(filtersMenu, "Variance Filter", variance_filter.gui)
    addAction(filtersMenu, "Median Filter", median_filter.gui)
    addAction(filtersMenu, "Fourier Filter", fourier_filter.gui)
    addAction(filtersMenu, "Difference Filter", difference_filter.gui)
    addAction(filtersMenu, "Boxcar Differential", boxcar_differential_filter.gui)
    addAction(filtersMenu, "Wavelet Filter", wavelet_filter.gui)
    addAction(filtersMenu, "Bilateral Filter", bilateral_filter.gui)

    g.menus.append(imageMenu)
    g.menus.append(processMenu)


# -*- coding: utf-8 -*-
import sys, os
import platform
import optparse
import warnings
import numpy as np
from .version import __version__
from .logger import logger
from .app.application import FlikaApplication

# for development purposes, add this if flika is not in your site-packages
# sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))



warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

def parse_arguments(argv):
    ''' Parses command line arguments for valid flika args

    :param argv: Arguments passed to program

    *Returns*
    A tuple of options, position arguments
    '''
    usage = """usage: %prog [FILE FILE...]

    # start a new session
    %prog

    # start a new session and load a file
    %prog image.tiff

    #start a new session with multiple files
    %prog image.tiff script.py

    #run a script
    %prog -x script.py

    #increase verbosity level
    %prog -v

    #run the test suite0

    %prog -t
    """
    parser = optparse.OptionParser(usage=usage, version=str(__version__))

    parser.add_option('-x', '--execute', action='store_true', dest='script',
                      help="Open file in script editor and run", default=False)
    parser.add_option('-t', '--test', action='store_true', dest='test',
                      help="Run test suite", default=False)
    parser.add_option('-v', '--verbose', action='store_true',
                      help="Increase the vebosity level", default=False)

    err_msg = verify(parser, argv)
    if err_msg:
        sys.stderr.write('\n%s\n' % err_msg)
        parser.print_help()
        sys.exit(1)

    return parser.parse_args(argv)

def verify(parser, argv):
    """ Check for input errors

    :param parser: OptionParser instance
    :param argv: Argument list
    :type argv: List of strings

    *Returns*
    An error message, or None
    """
    opts, args = parser.parse_args(argv)
    err_msg = None
    if opts.script and len(args) != 1:
        err_msg = "Must provide a script\n"

    return err_msg

def ipython_qt_event_loop_setup():
    try:
        __IPYTHON__
    except NameError:
        return #  If __IPYTHON__ is not defined, we are not in ipython
    else:
        print("Starting flika inside IPython")
        from IPython import get_ipython
        ipython = get_ipython()
        ipython.magic("gui qt")

def load_files(files):
    from .process.file_ import open_file
    for f in files:
        open_file(f)

def start_flika(files=[]):
    """Run a flika session and exit

    Parameters
    ----------
    files : list
        An optional list of data files to load.
 
    """
    fa = FlikaApplication()
    load_files(files)
    fa.start()
    ipython_qt_event_loop_setup()
    return fa

def exec_():
    fa = start_flika(sys.argv[1:])
    return fa.app.exec_()

def post_install():
    if platform.system() == 'Windows':
        print("Creating start menu shortcut...")
        import winshell
        icon_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'images', 'favicon.ico')
        flika_exe = os.path.join(sys.exec_prefix, 'Scripts', 'flika.exe')
        link_path = os.path.join(winshell.programs(), "flika.lnk")
        with winshell.shortcut(link_path) as link:
            link.path = flika_exe
            link.description = "flika"
            link.icon_location = (icon_path, 0)
        link_path = os.path.join(winshell.desktop(), "flika.lnk")
        with winshell.shortcut(link_path) as link:
            link.path = flika_exe
            link.description = "flika"
            link.icon_location = (icon_path, 0)

if __name__ == '__main__':
    start_flika(sys.argv[1:])


"""
def exec_(args=sys.argv):
    opt, args = parse_arguments(args[1:])

    if opt.verbose:
        logger.setLevel("INFO")

    logger.info("Input arguments: %s", sys.argv)

    start_flika(files=args)


def run(args=sys.argv):
    ''' open flika without running exec_. For debugging purposes
    '''
    opt, args = parse_arguments(args[1:])

    if opt.verbose:
        logger.setLevel("INFO")

    fa = FlikaApplication()
    fa.show()

    load_files(files=args)

    if 'PYCHARM_HOSTED' not in os.environ and 'SPYDER_SHELL_ID' not in os.environ:
        return fa.app.exec_()
"""




# -*- coding: utf-8 -*-
import os.path
import numpy as np
from qtpy import QtCore, QtGui, QtWidgets
import pyqtgraph as pg
import sys
import inspect
import markdown

from .. import global_vars as g
from .. import window
from ..utils.misc import save_file_gui

__all__ = ['MissingWindowError', 'WindowSelector', 'FileSelector', 'ColorSelector', 'SliderLabel', 'SliderLabelOdd', 'CheckBox', 'ComboBox', 'BaseDialog', 'BaseProcess', 'BaseProcess_noPriorWindow']
        
class MissingWindowError(Exception):
    def __init__(self, value):
        self.value = value
        g.m.statusBar().showMessage(value)
    def __str__(self):
        return repr(self.value)


class WindowSelector(QtWidgets.QWidget):
    """
    This widget is a button with a label.  Once you click the button, the widget waits for you to click a Window object.  Once you do, it sets self.window to be the window, and it sets the label to be the widget name.
    """
    valueChanged=QtCore.Signal()
    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        self.button = QtWidgets.QPushButton('Select Window')
        self.button.setCheckable(True)
        self.label = QtWidgets.QLabel('None')
        self.window = None
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.button)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)
        self.button.clicked.connect(self.buttonclicked)

    def buttonclicked(self):
        if self.button.isChecked() is False:
            g.m.setCurrentWindowSignal.sig.disconnect(self.setWindow)
        else:
            g.m.setCurrentWindowSignal.sig.connect(self.setWindow)

    def setWindow(self, window=None):
        if window is None:
            try:
                g.m.setCurrentWindowSignal.sig.disconnect(self.setWindow)
            except TypeError:
                pass
            self.window = g.currentWindow
        else:
            self.window = window
        self.button.setChecked(False)
        self.label.setText('...'+os.path.split(self.window.name)[-1][-20:])
        self.valueChanged.emit()
        self.parent().raise_()
    def value(self):
        return self.window
    def setValue(self, window):
        ''' This function is written to satify the requirement that all items have a setValue function to recall from settings the last set value. '''
        self.setWindow(window)


class FileSelector(QtWidgets.QWidget):
    """
    This widget is a button with a label.  Once you click the button, the widget waits for you to select a file to save.  Once you do, it sets self.filename and it sets the label.
    """
    valueChanged=QtCore.Signal()
    def __init__(self,filetypes='*.*'):
        QtWidgets.QWidget.__init__(self)
        self.button=QtWidgets.QPushButton('Select Filename')
        self.label=QtWidgets.QLabel('None')
        self.window=None
        self.layout=QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.button)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)
        self.button.clicked.connect(self.buttonclicked)
        self.filetypes = filetypes
        self.filename = ''
        
    def buttonclicked(self):
        prompt = 'testing fileSelector'
        self.filename = save_file_gui(prompt, filetypes=self.filetypes)
        self.label.setText('...'+os.path.split(self.filename)[-1][-20:])
        self.valueChanged.emit()

    def value(self):
        return self.filename

    def setValue(self, filename):
        self.filename = str(filename)
        self.label.setText('...' + os.path.split(self.filename)[-1][-20:])

def color_pixmap(color):
    pm = QtGui.QPixmap(15, 15)
    p = QtGui.QPainter(pm)
    b = QtGui.QBrush(QtGui.QColor(color))
    p.fillRect(-1, -1, 20, 20, b)
    return pm

class ColorSelector(QtWidgets.QWidget):
    """
    This widget is a button with a label.  Once you click the button, the widget waits for you to select a color.  Once you do, it sets self.color and it sets the label.
    """
    valueChanged=QtCore.Signal()
    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        self.button=QtWidgets.QPushButton('Select Color')
        self.label=QtWidgets.QLabel('None')
        self.window=None
        self.layout=QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.button)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)
        self.button.clicked.connect(self.buttonclicked)
        self._color=''
        self.colorDialog=QtWidgets.QColorDialog()
        self.colorDialog.colorSelected.connect(self.colorSelected)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = color
        self.button.setIcon(QtGui.QIcon(color_pixmap(color)))

    def buttonclicked(self):
        self.colorDialog.open()

    def value(self):
        return self._color
        
    def colorSelected(self, color):
        if color.isValid():
            self.color=color.name()
            self.label.setText(color.name())
            self.valueChanged.emit()


class SliderLabel(QtWidgets.QWidget):
    changeSignal=QtCore.Signal(int)
    def __init__(self,decimals=0): #decimals specifies the resolution of the slider.  0 means only integers,  1 means the tens place, etc.
        QtWidgets.QWidget.__init__(self)
        self.slider=QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.decimals=decimals
        if self.decimals<=0:
            self.label=QtWidgets.QSpinBox()
        else:
            self.label=QtWidgets.QDoubleSpinBox()
            self.label.setDecimals(self.decimals)
        self.layout=QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.slider)
        self.layout.addWidget(self.label)
        self.layout.setContentsMargins(0,0,0,0)
        self.setLayout(self.layout)
        self.slider.valueChanged.connect(lambda val: self.updateLabel(val/10**self.decimals))
        self.label.valueChanged.connect(self.updateSlider)
        self.valueChanged=self.label.valueChanged
    @QtCore.Slot(float)
    @QtCore.Slot(int)
    def updateSlider(self,value):
        self.slider.setValue(int(value*10**self.decimals))
    def updateLabel(self,value):
        self.label.setValue(value)
    def value(self):
        return self.label.value()
    def setRange(self,minn,maxx):
        self.slider.setRange(minn*10**self.decimals,maxx*10**self.decimals)
        self.label.setRange(minn,maxx)
    def setMinimum(self,minn):
        self.slider.setMinimum(minn*10**self.decimals)
        self.label.setMinimum(minn)
    def setMaximum(self,maxx):
        self.slider.setMaximum(maxx*10**self.decimals)
        self.label.setMaximum(maxx)
    def setValue(self,value):
        self.slider.setValue(value*10**self.decimals)
        self.label.setValue(value)
    def setSingleStep(self,value):
        self.label.setSingleStep(value)
    def setEnabled(self, en):
        self.label.setEnabled(en)
        self.slider.setEnabled(en)


class SliderLabelOdd(SliderLabel):
    '''This is a modified SliderLabel class that forces the user to only choose odd numbers.'''
    def __init__(self):
        SliderLabel.__init__(self,decimals=0)
    @QtCore.Slot(int)
    def updateSlider(self,value):
        value=int(value)
        if value%2==0: #if value is even
            value+=1
        self.slider.setValue(value)
    def updateLabel(self,value):
        value=int(value)
        if value%2==0: #if value is even
            value+=1
        self.label.setValue(value)


class CheckBox(QtWidgets.QCheckBox):
    ''' I overwrote the QCheckBox class so that every graphical element has the method 'setValue'
    '''
    def __init__(self,parent=None):
        QtWidgets.QCheckBox.__init__(self,parent)
    def setValue(self,value):
        self.setChecked(value)

class ComboBox(QtWidgets.QComboBox):
    ''' I overwrote the QComboBox class so that every graphical element has the method 'setValue'
    '''
    def __init__(self, parent=None):
        QtWidgets.QComboBox.__init__(self, parent)
    def setValue(self, value):
        if isinstance(value, str):
            idx = self.findText(value)
        else:
            idx = self.findData(value)
        if idx != -1:
            self.setCurrentIndex(idx)

class BaseDialog(QtWidgets.QDialog):
    changeSignal=QtCore.Signal()
    closeSignal=QtCore.Signal()
    def __init__(self, items, title, docstring, parent=None):
        QtWidgets.QDialog.__init__(self)
        self.parent = parent
        self.setWindowTitle(title)
        self.formlayout = QtWidgets.QFormLayout()
        self.formlayout.setLabelAlignment(QtCore.Qt.AlignRight)
        self.docstring = None
        self.items = items
        self.setupitems()
        self.connectToChangeSignal()
        self.bbox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.bbox.accepted.connect(self.accept)
        self.bbox.rejected.connect(self.reject)
        self._init_docstring(docstring)
        self.layout = QtWidgets.QVBoxLayout()
        if docstring is not None:
            self.layout.addWidget(self.docstring)
        self.layout.addLayout(self.formlayout)
        self.layout.addWidget(self.bbox)
        self.setLayout(self.layout)
        self.changeSignal.connect(self.updateValues)
        self.updateValues()

    def _init_docstring(self, docstring):
        if docstring is None:
            return None
        self.docstring = QtWidgets.QTextBrowser()
        self.docstring.setOpenExternalLinks(True)
        docstring = markdown.markdown(docstring)
        self.docstring.setHtml(docstring)
        #css = """ """
        #self.docstring.setStyleSheet(css)
        self.docstring.setWordWrapMode(0)

    def setupitems(self):
        for item in self.items:
            self.formlayout.addRow(item['string'], item['object'])
        # Get the old vals from settings
        if self.parent is not None:
            name = self.parent.__name__
            if g.settings['baseprocesses'] is None:
                g.settings['baseprocesses'] = dict()
            if name not in g.settings['baseprocesses']:
                settings = self.parent.get_init_settings_dict()
                g.settings['baseprocesses'][name] = settings
            else:
                settings = g.settings['baseprocesses'][name]
            for item in self.items:
                if item['name'] in settings and not isinstance(item['object'], WindowSelector):
                    item['object'].setValue(settings[item['name']])

    def connectToChangeSignal(self):
        for item in self.items:
            methods=[method for method in dir(item['object']) if callable(getattr(item['object'], method))]
            if 'valueChanged' in methods:
                item['object'].valueChanged.connect(self.changeSignal)
            elif 'stateChanged' in methods:
                item['object'].stateChanged.connect(self.changeSignal)
            elif 'currentIndexChanged' in methods:
                item['object'].currentIndexChanged.connect(self.changeSignal)

    def updateValues(self): # copy values from gui into the 'item' dictionary
        for item in self.items:
            methods=[method for method in dir(item['object']) if callable(getattr(item['object'], method))]
            if 'value' in methods:
                item['value']=item['object'].value()
            elif 'currentText' in methods:
                item['value']=item['object'].currentText()
            elif 'isChecked' in methods:
                item['value']=item['object'].isChecked()
    def closeEvent(self,ev):
        self.closeSignal.emit()


def convert_to_string(item):
    if isinstance(item, str):
        return "'{}'".format(item)
    else:
        return str(item)


class BaseProcess(object):
    """Subclass BaseProcess when writing your own process. Why would you want to use BaseProcess? 


    """
    def __init__(self):
        self.noPriorWindow = False
        self.__name__=self.__class__.__name__.lower()
        self.items=[]

    def getValue(self,name):
        return [i['value'] for i in self.items if i['name']==name][0]

    def get_init_settings_dict(self):
        return dict() #this function needs to be overwritten by every subclass

    def start(self, keepSourceWindow):
        frame = inspect.getouterframes(inspect.currentframe())[1][0]
        args, _, _, values = inspect.getargvalues(frame)
        funcname = self.__name__
        self.command = funcname+'('+', '.join([i+'='+convert_to_string(values[i]) for i in args if i!='self'])+')'
        g.m.statusBar().showMessage('Running function {}...'.format(self.__name__))
        self.keepSourceWindow = keepSourceWindow
        self.oldwindow = g.currentWindow
        if self.oldwindow is None:
            raise(MissingWindowError("You cannot execute '{}' without selecting a window first.".format(self.__name__)))
        self.tif = self.oldwindow.image
        self.oldname = self.oldwindow.name

    def end(self):
        if not hasattr(self, 'newtif') or self.newtif is None:
            self.oldwindow.reset()
            return
        commands=self.oldwindow.commands[:]
        if hasattr(self, 'command'):
            commands.append(self.command)
        newWindow=window.Window(self.newtif,str(self.newname),self.oldwindow.filename,commands,self.oldwindow.metadata)
        if self.keepSourceWindow is False:
            self.oldwindow.close()
        else:
            self.oldwindow.reset()
        if np.max(self.newtif)==1 and np.min(self.newtif)==0: #if the array is boolean
            newWindow.imageview.setLevels(-.1,1.1)
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        del self.tif
        del self.newtif
        return newWindow

    def gui(self):
        self.ui=BaseDialog(self.items,self.__name__,self.__doc__, self)
        if hasattr(self, '__url__'):
            self.ui.bbox.addButton(QtWidgets.QDialogButtonBox.Help)
            self.ui.bbox.helpRequested.connect(lambda : QtWidgets.QDesktopServices.openUrl(QtCore.QUrl(self.__url__)))
        self.proxy= pg.SignalProxy(self.ui.changeSignal,rateLimit=60, slot=self.preview)
        if g.currentWindow is not None:
            self.ui.rejected.connect(g.currentWindow.reset)
        self.ui.closeSignal.connect(self.ui.rejected.emit)
        self.ui.accepted.connect(self.call_from_gui)
        self.ui.show()
        g.dialogs.append(self.ui)
        return True

    def gui_reset(self):
        self.items=[]

    def call_from_gui(self):
        varnames = [i for i in inspect.getargspec(self.__call__)[0] if i != 'self' and i != 'keepSourceWindow']
        try:
            args = [self.getValue(name) for name in varnames]
        except IndexError as err:
            msg = "IndexError in {}: {}".format(self.__name__, varnames)
            msg += str(err)
            g.alert(msg)
        newsettings = dict()
        for name in varnames:
            value = self.getValue(name)
            if not isinstance(value, window.Window): # cannot save window objects using pickle
                newsettings[name] = value
        g.settings['baseprocesses'][self.__name__] = newsettings
        g.settings.save()
        try:
            if self.noPriorWindow:
                self.__call__(*args)
            else:
                self.__call__(*args, keepSourceWindow=True)
        except MemoryError as err:
            msg = 'There was a memory error in {}. Close other programs and try again.'.format(self.__name__)
            msg += str(err)
            g.alert(msg)
    def preview(self):
        pass


class BaseProcess_noPriorWindow(BaseProcess):

    def __init__(self):
        super().__init__()
        self.noPriorWindow = True

    def start(self):
        frame = inspect.getouterframes(inspect.currentframe())[1][0]
        args, _, _, values = inspect.getargvalues(frame)
        funcname = self.__name__
        self.command = funcname+'('+', '.join([i+'='+convert_to_string(values[i]) for i in args if i!='self'])+')'
        g.m.statusBar().showMessage('Performing {}...'.format(self.__name__))
        
    def end(self):
        commands = [self.command]
        newWindow = window.Window(self.newtif,str(self.newname),commands=commands)
        if np.max(self.newtif) == 1 and np.min(self.newtif) == 0: #if the array is boolean
            newWindow.imageview.setLevels(-.1,1.1)
        g.m.statusBar().showMessage('Finished with {}.'.format(self.__name__))
        del self.newtif
        return newWindow





































